/*
 * XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
 * SOLELY FOR USE IN DEVELOPING PROGRAMS AND SOLUTIONS FOR
 * XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION
 * AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION
 * OR STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS
 * IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
 * AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
 * FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
 * WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
 * IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
 * REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
 * INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE.
 *     
 * (c) Copyright 2006 Xilinx, Inc.
 * All rights reserved.
 *
 */
 
/*
 * Modification history:
 *    01/2006 - Rob Armstrong - Xilinx, Inc.
 *            - Initial File Version
 */
 
#include "ethernet.h" 
#include "serial_io.h"
 
int processPacket(unsigned char *rxFrame,
                   unsigned char *macAddress,
                   unsigned char *ipAddress,
                   unsigned long receiveLength)
{
    int result;
                   
    // Determine whether or not the Ethernet frame is targeted to us
    if((isBroadcast(rxFrame) == TRUE) || (isUnicastMatch(rxFrame, macAddress) == TRUE)) {
        if(isArp(rxFrame)) {
            result = processArp(rxFrame, macAddress, ipAddress);
            if (result == ARP_DIFFERENT_TARGET) {
                xil_printf("ARP request received for different target\r\n");
                return NO_RETRANSMIT;
            }
            else if(result == ARP_NOT_REQUEST) {
                xil_printf("ARP packet received but not a valid request\r\n");
                return NO_RETRANSMIT;
            }
            else {
                xil_printf("ARP packet received and response generated\r\n");
                return RETRANSMIT;
            }
        }
        else if(isPing(rxFrame) == TRUE) {
            processPing(rxFrame, macAddress, ipAddress, receiveLength);
            xil_printf("ICMP Echo (Ping) request received and response generated\r\n");
            return RETRANSMIT;                 
        }
        else {
            xil_printf("Received unknown packet type - ignoring\r\n");
            return NO_RETRANSMIT;
        }
    }               
}

/* int isBroadcast(unsigned char *rxFrame)
 * 
 * Compares the destination mac address to FF:FF:FF:FF:FF:FF to determine
 * whether or not the frame is a broadcast frame
 */
int isBroadcast(unsigned char *rxFrame)
{
    int loop;
    unsigned char *macPointer;
    
    // Loop through the destination MAC address to check that all bytes are
    // set to 0xFF. If they are, return TRUE, otherwise return FALSE.
    macPointer = rxFrame + DEST_MAC_OFFSET;
    for(loop = 0; loop < 6; loop++) {
        if(*macPointer != 0xFF) {
            return FALSE;
        }
        macPointer++;
    }
    return TRUE;
}

/* int isUnicastMatch(unsigned char *rxFrame, unsigned char *macAddress);
 * 
 * Compares the destination mac address to the MAC address set by the user to 
 * determine whether or not the frame is a unicast frame targeting this MAC
 */
int isUnicastMatch(unsigned char *rxFrame, unsigned char *macAddress)
{
    int loop;
    unsigned char *macPointer;
    
    // Loop through the destination MAC address to check that all bytes are
    // equal to the preset MAC values. If they are, return TRUE, otherwise
    // return FALSE.
    macPointer = rxFrame + DEST_MAC_OFFSET;
    for(loop = 0; loop < 6; loop++) {
        if(*(macPointer + loop) != *(macAddress + loop)) {
            return FALSE;
        }
    }
    return TRUE;
}

/* int isArp(unsigned char *rxFrame)
 * 
 * Compares the Length/Type field with 0x0806 to see if this is an ARP packet.
 * Returns TRUE if it is, and otherwise returns FALSE.
 */
int isArp(unsigned char *rxFrame)
{
    unsigned char *ltPointer;
    
    ltPointer = rxFrame + LENGTHTYPE_OFFSET;
    if((*ltPointer == 0x08) && (*(ltPointer + 1) == 0x06)) {
        return TRUE;
    }
    return FALSE;
}

/* int processArp(unsigned char *rxFrame,
 *                unsigned char *macAddress,
 *                unsigned char *ipAddress)
 * 
 * Process the ARP packet. If it is targeted to the IP address set by the
 * user, format and send the response.
 */
int processArp(unsigned char *rxFrame,
               unsigned char *macAddress,
               unsigned char *ipAddress)
{
    unsigned char *framePointer;
    int loop;
    unsigned char swapBuffer[10];
    
    // Check to see if the IP address targeted in the ARP request matches
    // the IP address set by the user
    framePointer = rxFrame + ARP_TARGET_IP_OFFSET;
    for(loop = 0; loop < 4; loop++) {
        if(*framePointer++ != *(ipAddress + loop)) {
            return ARP_DIFFERENT_TARGET;
        }
    }
    
    // Check the Opcode in the IP frame to make sure that this is an ARP
    // request - we shouldn't see an ARP broadcast packet with a response and
    // an IP address matching ours, but we're required to check
    framePointer = rxFrame + ARP_OPCODE_OFFSET;
    if((*framePointer != 0x00) && (*(framePointer + 1) != 0x01)) {
        return ARP_NOT_REQUEST;
    }
    
    // Swap the source and destination MACs, and set the source MAC to the MAC
    // address that was set by the user.
    swapSourceDestMac(rxFrame);
    memcpy(rxFrame + SOURCE_MAC_OFFSET, macAddress, 6);
    
    // Swap the source and target identification fields in the ARP packet
    memcpy(swapBuffer, rxFrame + ARP_SOURCE_MAC_OFFSET, 10);
    memcpy(rxFrame + ARP_SOURCE_MAC_OFFSET, rxFrame + ARP_DEST_MAC_OFFSET, 10);
    memcpy(rxFrame + ARP_DEST_MAC_OFFSET, swapBuffer, 10);
    
    // Insert the the MAC address set by the user into the ARP Source field
    memcpy(rxFrame + ARP_SOURCE_MAC_OFFSET, macAddress, 6);
    
    // Set the opcode from a request to a reply
    framePointer = rxFrame + ARP_OPCODE_OFFSET + 1;
    *framePointer = ARP_OPCODE_REPLY;
    return ARP_RESPONSE_OK;
}

/* void swapSourceDestMac(unsigned char *rxFrame)
 * 
 * Swap the source and destination MAC addresses in an ethernet frame
 */
void swapSourceDestMac(unsigned char *rxFrame)
{
    unsigned char macBuffer[6];
    
    // Copy the destination MAC address to the address buffer, the source MAC
    // to the destination MAC field, and the destination MAC from the buffer
    // back into the source MAC field
    memcpy(macBuffer, rxFrame + DEST_MAC_OFFSET, 6);
    memcpy(rxFrame + DEST_MAC_OFFSET, rxFrame + SOURCE_MAC_OFFSET, 6);
    memcpy(rxFrame + SOURCE_MAC_OFFSET, macBuffer, 6);
}

/* int isPing(unsigned char *rxFrame)
 * 
 * Compares the Length/Type field with 0x0800 to see if this is an IP packet.
 * If so, check the protocol and type fields to determine whether or not this
 * is an ICMP Echo request
 */
int isPing(unsigned char *rxFrame)
{
    unsigned char *framePointer;
    
    // Check the Length/Type field for the proper value (0x0800)
    framePointer = rxFrame + LENGTHTYPE_OFFSET;
    if((*framePointer == 0x08) && (*(framePointer + 1) == 0x00)) {
        
        // Check the Protocol field for the ICMP protocol in the IP header
        framePointer = rxFrame + IP_PROTOCOL_OFFSET;
        if(*framePointer == ICMP) {
            
            // Check the ICMP type field to see if this is an echo request
            framePointer = rxFrame + IP_ICMP_TYPE_OFFSET;
            
            if(*framePointer == ECHO_REQUEST) {
                return TRUE;
            }
        }
    }
    // If the above tests don't all succeed, this isn't an ICMP echo request
    return FALSE;
}

/* void processPing(unsigned char *rxFrame,
 *                  unsigned char *macAddress,
 *                  unsigned char *ipAddress,
 *                  unsigned long receiveLength)
 * 
 * Process the ICMP Echo packet. Format and send the response.
 */
void processPing(unsigned char *rxFrame,
                unsigned char *macAddress,
                unsigned char *ipAddress,
                unsigned long receiveLength)
{
    unsigned char *framePointer;
    int loop;
    unsigned char swapBuffer[10];
    unsigned short checksum;
    
    // Swap the source and destination MACs in the packet to return the ping
    swapSourceDestMac(rxFrame);
    
    // Set the header's identification field to an appropriate value (must not
    // equal the ID field used by the sender). To do this simply we will increment
    // the ID field of the sent packet by 1
    framePointer = rxFrame + IP_ID_OFFSET;
    if(*(framePointer + 1) < 0xFF) {
        *(framePointer + 1) += 1;
    }
    else {
        *framePointer += 1;
        *(framePointer + 1) = 0;
    }
    
    // Swap the source and destination IP fields in the packet header
    memcpy(swapBuffer, rxFrame + IP_DEST_IP_OFFSET, 4);
    memcpy(rxFrame + IP_DEST_IP_OFFSET, rxFrame + IP_SOURCE_IP_OFFSET, 4);
    memcpy(rxFrame + IP_SOURCE_IP_OFFSET, swapBuffer, 4);
    
    // Set the checksum for the IP header. This is calculated from the end of the
    // Length/Type field down to the end of the destination IP.
    framePointer = rxFrame + IP_HEADER_CHECKSUM_OFFSET;
    *framePointer = 0;
    *(framePointer + 1) = 0;
    checksum = calcChecksum((rxFrame + IP_HEADER_OFFSET), IP_HEADER_LENGTH);
    *framePointer = (unsigned char)((checksum & 0xFF00) >> 8);
    *(framePointer + 1) = (unsigned char)(checksum & 0x00FF);
    
    // Set the ICMP message type to 0 (ICMP Echo reply)
    framePointer = rxFrame + IP_ICMP_TYPE_OFFSET;
    *framePointer = ECHO_REPLY;
    
    // Set the checksum for the ICMP data field
    framePointer = rxFrame + IP_ICMP_CHECKSUM_OFFSET;
    *framePointer = 0;
    *(framePointer + 1) = 0;
    checksum = calcChecksum((rxFrame + IP_DATA_OFFSET), 
                            (receiveLength - IP_DATA_OFFSET - CRC_LENGTH));
    *framePointer = (unsigned char)((checksum & 0xFF00) >> 8);
    *(framePointer + 1) = (unsigned char)(checksum & 0x00FF);   
    
    
}

/* unsigned short calcChecksum(unsigned char *data,
 *                             int length);
 *
 * Calculate the 16-bit checksum used in the IP and ICMP headers
 */
unsigned short calcChecksum(unsigned char *data, int length)
{
   unsigned int checksum;
   unsigned short w16;
   int loop;
   
   checksum = 0;
   w16 = 0;

   for (loop = 0; loop < length; loop += 2) {
      w16 = ((data[loop] << 8) & 0xFF00) + (data[loop + 1] & 0x00FF);
      checksum += (unsigned int) w16;
   }
  
   while (checksum >> 16) { 
      checksum = (checksum & 0xFFFF) + (checksum >> 16);
   }
  
   return (unsigned short) (~checksum);
}
